

float sigmoid(__private float x)
{
    return 1 / (1 + exp(-x)); //native_exp
}


void propagate_layer_local(__local float *input,
                           __local float *weights,
                           __local float *output,
                           int previous_layer_size,
                           int layer_size)
{
    for(int i=0;i<layer_size;i++)
    {
        float temp_output = weights[i*(previous_layer_size+1)];
        for(int j=0;j<previous_layer_size;j++)
        {
            //temp_output = fma(input[j],weights[i*(previous_layer_size+1)+j+1],temp_output);
            //temp_output = mad(input[j],weights[i*(previous_layer_size+1)+j+1],temp_output);
            temp_output += input[j]*weights[i*(previous_layer_size+1)+j+1];
        }
        output[i] = sigmoid(temp_output);
    }
}


// output and temp_output have to be the size of the widest layer. Both memory locations are used in calculations.
void simulate_neural_network_local(__global float *input,
                                   __local float *weights,
                                   __local float *output,
                                   int num_layers,
                                   __constant int *layers,
                                   __constant float *out_ranges,
                                   __local float *temp_output)
{
    int input_size = layers[0];
    // copy input to temp
    if(num_layers%2==0)
        for(int i=0;i<input_size;i++)
            temp_output[i] = input[i];
    else //(num_layers%2=1)
        for(int i=0;i<input_size;i++)
            output[i] = input[i];

    if(num_layers%2==0) propagate_layer_local(temp_output, weights, output,layers[0],layers[1]);
    else propagate_layer_local(output, weights, temp_output,layers[0],layers[1]);
    int weights_index = layers[1]*(layers[0]+1);

    for(int i=2;i<num_layers;i++)
    {
        if((num_layers-i)%2==1) propagate_layer_local(temp_output, &weights[weights_index], output,layers[i-1],layers[i]);
        else propagate_layer_local(output, &weights[weights_index], temp_output,layers[i-1],layers[i]);
        weights_index += layers[i]*(layers[i-1]+1);
    }

    // now scale the output to out_ranges
    int outputSize = layers[num_layers-1];
    for(int i=0;i<outputSize;i++)
        output[i] = output[i] * (out_ranges[2*i+1] - out_ranges[2*i]) + out_ranges[2*i];

    //result is now in output
}




__kernel void simulate_neural_networks_local(__global float *inputs,
                                             __global float *weights,
                                             __global float *outputs,
                                             int num_layers,
                                             __constant int *layers,
                                             __constant float *out_ranges,
                                             int number_of_networks,
                                             int number_of_inputs,
                                             int num_weights,
                                             int widest_layer,
                                             __local float *sh_weights,
                                             __local float *sh_output,
                                             __local float *sh_temp_output)
{


    int idx = get_local_id(1);
    int test_idx = get_global_id(1); //blockIdx.y * blockDim.x + threadIdx.x;
    int nn_idx = get_group_id(0); //blockIdx.x;


    __local float *sh_output_ = &sh_output[idx*widest_layer];
    __local float *sh_temp_output_ = &sh_temp_output[idx*widest_layer];



    // load weights to local memory
    // loop je dosta brzi od async_work_group_copy (~x5)

    // event_t event = async_work_group_copy(sh_weights, &weights[num_weights*nn_idx], num_weights, 0);
    // wait_group_events(1,&event);


    int groupSize = get_local_size(1);
    for(int i=idx;i<num_weights;i+=groupSize)
        sh_weights[i] = weights[num_weights*nn_idx+i];
    barrier(CLK_LOCAL_MEM_FENCE);


    int input_size = layers[0];
    int output_size = layers[num_layers-1];

    if(test_idx < number_of_inputs)
    //if(test_idx < number_of_inputs && nn_idx < number_of_networks)
    {
        simulate_neural_network_local(&inputs[test_idx*input_size], sh_weights, sh_output_, num_layers, layers, out_ranges, sh_temp_output_);
        for(int i=0;i<output_size;i++)
            outputs[nn_idx*number_of_inputs*output_size+output_size*test_idx + i] = sh_output_[i];
    }
}


// private memory version
#define PRIVATE_MEM_SIZE 1000
// Before compiling "[PRIVATE_MEM_SIZE]" needs to be replaced with the size of the widest layer.

void propagate_layer_private(__private float *input,
                             __global float *weights,
                             __private float *output,
                             int previous_layer_size,
                             int layer_size)
{
    for(int i=0;i<layer_size;i++)
    {
        float temp_output = weights[i*(previous_layer_size+1)]; // bias
        for(int j=0;j<previous_layer_size;j++)
        {
            //temp_output = fma(input[j],weights[i*(previous_layer_size+1)+j+1],temp_output);
            //temp_output = mad(input[j],weights[i*(previous_layer_size+1)+j+1],temp_output);
            temp_output += input[j]*weights[i*(previous_layer_size+1)+j+1];
        }
        output[i] = sigmoid(temp_output);
    }
}


// output and temp_output have to be the size of the widest layer. Both memory locations are used in calculations.
void simulate_neural_network_private(__global float *input,
                                     __global float *weights,
                                     __private float *output,
                                     int num_layers,
                                     __constant int *layers,
                                     __constant float *out_ranges,
                                     __private float *temp_output)
{
    int input_size = layers[0];
    // copy input to temp
    if(num_layers%2==0)
        for(int i=0;i<input_size;i++)
            temp_output[i] = input[i];
    else //(num_layers%2=1)
        for(int i=0;i<input_size;i++)
            output[i] = input[i];

    if(num_layers%2==0) propagate_layer_private(temp_output, weights, output,layers[0],layers[1]);
    else propagate_layer_private(output, weights, temp_output,layers[0],layers[1]);
    int weights_index = layers[1]*(layers[0]+1);

    for(int i=2;i<num_layers;i++)
    {
        if((num_layers-i)%2==1) propagate_layer_private(temp_output, &weights[weights_index], output,layers[i-1],layers[i]);
        else propagate_layer_private(output, &weights[weights_index], temp_output,layers[i-1],layers[i]);
        weights_index += layers[i]*(layers[i-1]+1);
    }

    // now scale the output to [0,1]
    int outputSize = layers[num_layers-1];
    for(int i=0;i<outputSize;i++)
        output[i] = output[i] * (out_ranges[2*i+1] - out_ranges[2*i]) + out_ranges[2*i];

    //result is now in output
}




__kernel void simulate_neural_networks_private(__global float *inputs,
                                               __global float *weights,
                                               __global float *outputs,
                                               int num_layers,
                                               __constant int *layers,
                                               __constant float *out_ranges,
                                               int number_of_networks,
                                               int number_of_inputs,
                                               int num_weights)
{


    int idx = get_local_id(1);
    int test_idx = get_global_id(1);
    int nn_idx = get_group_id(0);

    __private float priv_output[PRIVATE_MEM_SIZE];
    __private float priv_temp_output[PRIVATE_MEM_SIZE];

    int input_size = layers[0];
    int output_size = layers[num_layers-1];

    if(test_idx < number_of_inputs)
    //if(test_idx < number_of_inputs && nn_idx < number_of_networks)
    {
        simulate_neural_network_private(&inputs[test_idx*input_size], &weights[num_weights*nn_idx], priv_output, num_layers, layers, out_ranges, priv_temp_output);
        for(int i=0;i<output_size;i++)
            outputs[nn_idx*number_of_inputs*output_size+output_size*test_idx + i] = priv_output[i];
    }
}



// Mean squared error
// Each group calculates MSE for one network. Each thread calculates partial sum and stores it in local memory,
// and one thread per group sums them all up and writes it to global memory.
__kernel void mean_squared_errors(__global float *outputs, __global float *targets, __global float *MSEs, int num_samples, int output_size, __local float *partial_sums)
{
    int idx = get_local_id(0);
    int groupSize = get_local_size(0);
    int netSize = num_samples * output_size; // size of all the outputs for one network
    int netIdx = get_group_id(0);
    __global float *outputs_ = &outputs[netIdx * netSize];

    float pSum = 0;

    for(int i=idx;i<netSize;i+=groupSize)
    {
        float diff = targets[i] - outputs_[i];
        pSum += diff * diff;
    }

    // partial results are written to local memory
    partial_sums[idx] = pSum;
    barrier(CLK_LOCAL_MEM_FENCE);

    // one thread sums all the values
    if(idx == 0)
    {
        float total = 0;
        for(int i=0;i<groupSize;i++)
            total += partial_sums[i];
        MSEs[netIdx] = total/netSize;
    }
}


///  ------------- BP -------------

// Each work item is one neuron.
__kernel void layer_propagate_forward(__global float *input,
                                      int input_offset, // if input is from the samples
                                      __global float *weights,
                                      int weights_offset,
                                      __global float *output,
                                      __global float *all_outputs,
                                      int all_outputs_offset,
                                      int previous_layer_size,
                                      int layer_size,
                                      int isLastLayer,  // then scale output to out_ranges
                                      __constant float *out_ranges)
{
    int neuronIdx = get_global_id(0);
    if(neuronIdx < layer_size)
    {
        float temp_output = weights[weights_offset+neuronIdx*(previous_layer_size+1)];
        for(int j=0;j<previous_layer_size;j++)
        {
            //temp_output = fma(input[j],weights[i*(previous_layer_size+1)+j+1],temp_output);
            //temp_output = mad(input[j],weights[i*(previous_layer_size+1)+j+1],temp_output);
            temp_output += input[input_offset+j]*weights[weights_offset+neuronIdx*(previous_layer_size+1)+j+1];
        }


        temp_output = sigmoid(temp_output);

        // also store outputs in all_outputs
        all_outputs[all_outputs_offset + neuronIdx] = temp_output;

        // output is scaled to out_ranges
        if(isLastLayer)
            output[neuronIdx] = temp_output * (out_ranges[2*neuronIdx+1] - out_ranges[2*neuronIdx]) + out_ranges[2*neuronIdx];
        else
            output[neuronIdx] = temp_output;

        //printf("out=%f\n",temp_output);

    }
}

// Each work item is one neuron.
// next/previous means the same as in forward propagation
__kernel void layer_propagate_backwards(__global float *all_outputs,
                                        int all_outputs_offset, // output of the first neuron in the current layer
                                        __global float *prev_deltas, // deltas previously calculated (from the next layer)
                                        __global float *deltas,
                                        __global float *weights,
                                        int weights_offset, // weights of the first neuron in the current layer (one from layer_size)
                                        __global float *delta_weights, // offset is same as for weights
                                        int prev_layer_size, // layer-1 sam as forward propagation
                                        int layer_size, // size of the current layer
                                        int next_layer_size, // layer+1
                                        float learningRate,
                                        int isLastLayer,
                                        __global float *targets,
                                        int targets_offset,
                                        __constant float *out_ranges)
{
    int neuronIdx = get_global_id(0);

    if(neuronIdx < layer_size)
    {
        // calculate deltas
        float neuron_output = all_outputs[all_outputs_offset + neuronIdx];

        float delta_sum = 0; // or error if output layer
        if(isLastLayer)
        {
            // scale targets to range [0,1]
            float scaled_target = (targets[targets_offset + neuronIdx] - out_ranges[2*neuronIdx]) / (out_ranges[2*neuronIdx+1] - out_ranges[2*neuronIdx]);
            delta_sum = scaled_target - neuron_output; // output error

        }
        else
        {
            __global float *next_layer_weights = &weights[weights_offset + (prev_layer_size+1)*layer_size];
            for(int k=0;k<next_layer_size;k++) // k == neuron from layer+1
            {
                __global float *next_neuron_weights = &next_layer_weights[(layer_size+1)*k];
                delta_sum += prev_deltas[k] * next_neuron_weights[neuronIdx+1]; // +1 becouse 0 == bias
                //printf("prev deltas[%d] = %f\n",k,prev_deltas[k]);
            }
        }


        float delta = neuron_output * (1-neuron_output) * delta_sum;

        deltas[neuronIdx] = delta; // store for future use


        //calculate delta weights
        __global float *neuron_delta_weights = &delta_weights[weights_offset + neuronIdx * (prev_layer_size+1)]; // +1 because bias in not in size
        __global float *prev_layer_outputs = &all_outputs[all_outputs_offset - prev_layer_size];

        // bias
        //neuron_delta_weights[0] += learningRate * delta; // += version is if delta weights are added after all the samples
        neuron_delta_weights[0] = learningRate * delta;

        // connections to the previus layer
        for(int k=0;k<prev_layer_size;k++)
            //neuron_delta_weights[k+1] +=     int getLayerSize(int layer);learningRate * delta * prev_layer_outputs[k]; // += version is if delta weights are added after all the samples
            neuron_delta_weights[k+1] = learningRate * delta * prev_layer_outputs[k]; //delta_weights[0] is bias, so +1
    }
}


__kernel void apply_delta_weights(__global float *weights, __global float *delta_weights, int num_weights)
{
    int idx = get_global_id(0);
    if(idx < num_weights)
    {
        weights[idx] += delta_weights[idx];
    }
}
